import 'dart:collection';
import 'package:flutter/widgets.dart';
import 'package:get/src/instance/get_instance.dart';
import 'package:get/src/navigation/root/smart_management.dart';
import 'package:get/src/state_manager/rx/rx_interface.dart';

class GetxController extends DisposableInterface {
  final HashSet<StateSetter> _updaters = HashSet<StateSetter>();

  final HashMap<String, StateSetter> _updatersIds =
      HashMap<String, StateSetter>();

  /// Update GetBuilder with update();
  void update([List<String> ids, bool condition = true]) {
    if (!condition) return;
    (ids == null)
        ? _updaters.forEach((rs) => rs(() {}))
        : ids.forEach((element) {
            _updatersIds[element](() {});
          });
  }

  void addListener(StateSetter listener) {
    _updaters.add(listener);
  }

  void removeListener(StateSetter listener) {
    _updaters.remove(listener);
  }

  void addListenerId(String key, StateSetter listener) {
    _updatersIds[key] = listener;
  }

  void removeListenerId(String key) {
    _updatersIds.remove(key);
  }

  @override
  void onInit() async {}

  @override
  void onReady() async {}

  @override
  void onClose() async {}
}

class GetBuilder<T extends GetxController> extends StatefulWidget {
  final Widget Function(T) builder;
  final bool global;
  final String id;
  final String tag;
  final bool autoRemove;
  final bool assignId;
  final void Function(State state) initState, dispose, didChangeDependencies;
  final void Function(GetBuilder oldWidget, State state) didUpdateWidget;
  final T init;
  const GetBuilder({
    Key key,
    this.init,
    this.global = true,
    @required this.builder,
    this.autoRemove = true,
    this.assignId = false,
    this.initState,
    this.tag,
    this.dispose,
    this.id,
    this.didChangeDependencies,
    this.didUpdateWidget,
  })  : assert(builder != null),
        super(key: key);
  @override
  _GetBuilderState<T> createState() => _GetBuilderState<T>();
}

class _GetBuilderState<T extends GetxController> extends State<GetBuilder<T>> {
  GetxController controller;
  bool isCreator = false;
  @override
  void initState() {
    super.initState();

    if (widget.initState != null) widget.initState(this);
    if (widget.global) {
      final isPrepared = GetInstance().isPrepared<T>(tag: widget.tag);
      final isRegistred = GetInstance().isRegistred<T>(tag: widget.tag);

      if (isPrepared) {
        if (GetConfig.smartManagement != SmartManagement.keepFactory) {
          isCreator = true;
        }
        controller = GetInstance().find<T>(tag: widget.tag);
      } else if (isRegistred) {
        controller = GetInstance().find<T>(tag: widget.tag);
        isCreator = false;
      } else {
        controller = widget.init;
        isCreator = true;
        GetInstance().put<T>(controller, tag: widget.tag);
      }
    } else {
      controller = widget.init;
      isCreator = true;
      controller?.onStart();
    }

    if (widget.global &&
        GetConfig.smartManagement == SmartManagement.onlyBuilder) {
      controller?.onStart();
    }
    (widget.id == null)
        ? controller.addListener(setState)
        : controller.addListenerId(widget.id, setState);
  }

  @override
  void dispose() {
    super.dispose();
    if (widget.dispose != null) widget.dispose(this);
    if (isCreator || widget.assignId) {
      if (widget.autoRemove && GetInstance().isRegistred<T>(tag: widget.tag)) {
        (widget.id == null)
            ? controller.removeListener(setState)
            : controller.removeListenerId(widget.id);
        GetInstance().delete<T>(tag: widget.tag);
      }
    } else {
      (widget.id == null)
          ? controller.removeListener(setState)
          : controller.removeListenerId(widget.id);
    }
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    if (widget.didChangeDependencies != null) {
      widget.didChangeDependencies(this);
    }
  }

  @override
  void didUpdateWidget(GetBuilder oldWidget) {
    super.didUpdateWidget(oldWidget as GetBuilder<T>);
    if (widget.didUpdateWidget != null) widget.didUpdateWidget(oldWidget, this);
  }

  @override
  Widget build(BuildContext context) {
    return widget.builder(controller);
  }
}

class Updater {
  final StateSetter updater;
  final String id;
  const Updater({this.updater, this.id});
}

// typedef UpdaterBuilder = Updater Function();

// class GetBuilder<T extends GetxController> extends StatefulWidget {
//   final Widget Function(T) builder;
//   final bool global;
//   final String id;
//   final String tag;
//   final bool autoRemove;
//   final bool assignId;
//   final void Function(State state) initState, dispose, didChangeDependencies;
//   final void Function(GetBuilder oldWidget, State state) didUpdateWidget;
//   final T init;
//   const GetBuilder({
//     Key key,
//     this.init,
//     this.global = true,
//     @required this.builder,
//     this.autoRemove = true,
//     this.assignId = false,
//     this.initState,
//     this.tag,
//     this.dispose,
//     this.id,
//     this.didChangeDependencies,
//     this.didUpdateWidget,
//   })  : assert(builder != null),
//         super(key: key);
//   @override
//   _GetBuilderState<T> createState() => _GetBuilderState<T>();
// }

// class _GetBuilderState<T extends GetxController> extends State<GetBuilder<T>> {
//   GetxController controller;
//   bool isCreator = false;
//   @override
//   void initState() {
//     super.initState();

//     if (widget.initState != null) widget.initState(this);
//     if (widget.global) {
//       final isPrepared = GetInstance().isPrepared<T>(tag: widget.tag);
//       final isRegistred = GetInstance().isRegistred<T>(tag: widget.tag);

//       if (isPrepared) {
//         if (GetConfig.smartManagement != SmartManagement.keepFactory) {
//           isCreator = true;
//         }
//         controller = GetInstance().find<T>(tag: widget.tag);
//       } else if (isRegistred) {
//         controller = GetInstance().find<T>(tag: widget.tag);
//         isCreator = false;
//       } else {
//         controller = widget.init;
//         isCreator = true;
//         GetInstance().put<T>(controller, tag: widget.tag);
//       }
//     } else {
//       controller = widget.init;
//       isCreator = true;
//       controller?.onStart();
//     }

//     if (widget.global &&
//         GetConfig.smartManagement == SmartManagement.onlyBuilder) {
//       controller?.onStart();
//     }
//     (widget.id == null)
//         ? controller.addListener(setState)
//         : controller.addListenerId(widget.id, setState);
//   }

//   @override
//   void dispose() {
//     super.dispose();
//     if (widget.dispose != null) widget.dispose(this);
//     if (isCreator || widget.assignId) {
//       if (widget.autoRemove && GetInstance().isRegistred<T>(tag: widget.tag)) {
//         (widget.id == null)
//             ? controller.removeListener(setState)
//             : controller.removeListenerId(widget.id);
//         GetInstance().delete<T>(tag: widget.tag);
//       }
//     } else {
//       (widget.id == null)
//           ? controller.removeListener(setState)
//           : controller.removeListenerId(widget.id);
//     }
//   }

//   @override
//   void didChangeDependencies() {
//     super.didChangeDependencies();
//     if (widget.didChangeDependencies != null) {
//       widget.didChangeDependencies(this);
//     }
//   }

//   @override
//   void didUpdateWidget(GetBuilder oldWidget) {
//     super.didUpdateWidget(oldWidget as GetBuilder<T>);
//     if (widget.didUpdateWidget != null) widget.didUpdateWidget(oldWidget, this);
//   }

//   Widget get notifyChilds {
//     final observer = Value.setter;
//     Value.setter = setState;
//     final result = widget.builder(controller);
//     Value.setter = observer;
//     return result;
//   }

//   @override
//   Widget build(BuildContext context) => notifyChilds;
// }

// abstract class ValueInterface<T> extends GetxController {}

// class Value<T> extends ValueInterface<T> {
//   Value([this._value]);
//   T _value;

//   T get value {
//     if (setter != null) {
//       addListener(setter);
//     }

//     return _value;
//   }

//   static StateSetter setter;

//   set value(T newValue) {
//     if (_value == newValue) return;
//     _value = newValue;
//     update();
//   }
// }
